//|------------------------------------------------------------------+
//|                                                     OLAPstat.mqh |
//|                                    Copyright (c) 2019, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |

//| This is a simple wrapper for OLAP classes to be embedded into    |
//| an expert advidser for analysis of single pass trading history   |
//| in the tester.                                                   |

//| OLAP code (OLAPcube.mqh) is described in                         |
//|                            https://www.mql5.com/en/articles/6602 |
//|------------------------------------------------------------------+


#include <OLAP/OLAPcube.mqh>

// custom selector example
class DaysRangeSelector: public DateTimeSelector<TRADE_RECORD_FIELDS>
{
  public:
    DaysRangeSelector(const int n): DateTimeSelector<TRADE_RECORD_FIELDS>(FIELD_DURATION, n)
    {
      _typename = typename(this);
    }
    
    virtual bool select(const Record *r, int &index) const override
    {
      double d = r.get(selector);
      int days = (int)(d / (60 * 60 * 24));
      index = MathMin(days, granularity - 1);
      return true;
    }
    
    virtual string getLabel(const int index) const override
    {
      return index < granularity - 1 ? ((index < 10 ? " ": "") + (string)index + "D") : ((string)index + "D+");
    }
};


// Main wrapper class. Should be instantiated in EA. Find usage example after the class.

class OLAPStats
{
  protected:
    AGGREGATORS AggregatorType;
    TRADE_RECORD_FIELDS AggregatorField;
    SORT_BY SortBy;
  
    int selectorCount;
    SELECTORS selectorArray[];
    TRADE_RECORD_FIELDS selectorField[];
    
    SELECTORS Filter1;
    TRADE_RECORD_FIELDS Filter1Field;
    double Filter1value1;
    double Filter1value2;
    
    void setDefaults()
    {
      AggregatorType = AGGREGATOR_SUM;
      AggregatorField = FIELD_PROFIT_AMOUNT;
      SortBy = SORT_BY_NONE;
      
      Filter1 = SELECTOR_NONE;
      Filter1Field = FIELD_NONE;
    }

  public:
    OLAPStats(const SELECTORS s1, const TRADE_RECORD_FIELDS f1 = FIELD_NONE)
    {
      selectorCount = 1;
      ArrayResize(selectorArray, selectorCount);
      ArrayResize(selectorField, selectorCount);
      selectorArray[0] = s1;
      selectorField[0] = f1;
      setDefaults();
    }

    OLAPStats(const SELECTORS s1, const TRADE_RECORD_FIELDS f1, const SELECTORS s2, const TRADE_RECORD_FIELDS f2 = FIELD_NONE)
    {
      selectorCount = 2;
      ArrayResize(selectorArray, selectorCount);
      ArrayResize(selectorField, selectorCount);
      selectorArray[0] = s1;
      selectorField[0] = f1;
      selectorArray[1] = s2;
      selectorField[1] = f2;
      setDefaults();
    }

    OLAPStats(const SELECTORS s1, const TRADE_RECORD_FIELDS f1, const SELECTORS s2, const TRADE_RECORD_FIELDS f2, const SELECTORS s3, const TRADE_RECORD_FIELDS f3 = FIELD_NONE)
    {
      selectorCount = 3;
      ArrayResize(selectorArray, selectorCount);
      ArrayResize(selectorField, selectorCount);
      selectorArray[0] = s1;
      selectorField[0] = f1;
      selectorArray[1] = s2;
      selectorField[1] = f2;
      selectorArray[2] = s3;
      selectorField[2] = f3;
      setDefaults();
    }
    
    void setAggregator(const AGGREGATORS type, const TRADE_RECORD_FIELDS field = FIELD_NONE)
    {
      AggregatorType = type;
      AggregatorField = field;
    }
    
    void setSorting(const SORT_BY sort)
    {
      SortBy = sort;
    }
    
    void setFilter(const SELECTORS s, const TRADE_RECORD_FIELDS f, const double v1 = 0, const double v2 = 0)
    {
      if(Filter1 == SELECTOR_NONE)
      {
        ArrayResize(selectorArray, selectorCount + 1);
        ArrayResize(selectorField, selectorCount + 1);
      }
      Filter1 = s;
      Filter1Field = f;
      selectorArray[selectorCount] = s;
      selectorField[selectorCount] = f;
      Filter1value1 = v1;
      Filter1value2 = v2;
    }
  
    Selector<TRADE_RECORD_FIELDS> *createSelector(int i)
    {
      switch(selectorArray[i])
      {
        case SELECTOR_TYPE:
          return new TypeSelector();
        case SELECTOR_SYMBOL:
          return new SymbolSelector();
        case SELECTOR_SERIAL:
          return new SerialNumberSelector();
        case SELECTOR_MAGIC:
          return new MagicSelector();
        case SELECTOR_PROFITABLE:
          return new ProfitableSelector();
        case SELECTOR_DURATION:
          return new DaysRangeSelector(15); // up to 14 days
        case SELECTOR_WEEKDAY:
          return selectorField[i] != FIELD_NONE ? new WeekDaySelector(selectorField[i]) : NULL;
        case SELECTOR_DAYHOUR:
          return selectorField[i] != FIELD_NONE ? new DayHourSelector(selectorField[i]) : NULL;
        case SELECTOR_HOURMINUTE:
          return selectorField[i] != FIELD_NONE ? new DayHourSelector(selectorField[i]) : NULL;
        case SELECTOR_SCALAR:
          return selectorField[i] != FIELD_NONE ? new TradeSelector(selectorField[i]) : NULL;
        case SELECTOR_QUANTS:
          return selectorField[i] != FIELD_NONE ? new QuantizationSelector(selectorField[i]) : NULL;
      }
      return NULL;
    }
  
    void process(const Display *display = NULL)
    {
      if(selectorCount == 0)
      {
        Print("No selectors. Setup at least one of them.");
        return;
      }
      HistoryDataAdapter<CustomTradeRecord> history;
      DataAdapter *adapter = &history;
      
      Analyst<TRADE_RECORD_FIELDS> *analyst;
      
      Selector<TRADE_RECORD_FIELDS> *selectors[];
      ArrayResize(selectors, selectorCount);
      
      for(int i = 0; i < selectorCount; i++)
      {
        selectors[i] = createSelector(i);
        if(selectors[i] == NULL)
        {
          Print("Selector ", i, " is empty. Setup selectors successively (don't leave a hole in-between), specify a field when required");
          return;
        }
      }

      Filter<TRADE_RECORD_FIELDS> *filters[];
      if(Filter1 != SELECTOR_NONE)
      {
        ArrayResize(filters, 1);
        Selector<TRADE_RECORD_FIELDS> *filterSelector = createSelector(selectorCount);
        if(Filter1value1 != Filter1value2)
        {
          filters[0] = new FilterRange<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1, Filter1value2);
        }
        else
        {
          filters[0] = new Filter<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1);
        }
      }
      
      Aggregator<TRADE_RECORD_FIELDS> *aggregator;
      
      // MQL does not support a 'class info' metaclass.
      // Otherwise we could use an array of classes instead of the switch
      switch(AggregatorType)
      {
        case AGGREGATOR_SUM:
          aggregator = new SumAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_AVERAGE:
          aggregator = new AverageAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_MAX:
          aggregator = new MaxAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_MIN:
          aggregator = new MinAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_COUNT:
          aggregator = new CountAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_PROFITFACTOR:
          aggregator = new ProfitFactorAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_PROGRESSIVE:
          aggregator = new ProgressiveTotalAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
        case AGGREGATOR_IDENTITY:
          aggregator = new IdentityAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
          break;
      }
      
      LogDisplay stdDisplay(10, 3);
      Display *ptr = (Display *)display;
      if(ptr == NULL) ptr = &stdDisplay;
      
      analyst = new Analyst<TRADE_RECORD_FIELDS>(adapter, aggregator, ptr);
      
      analyst.acquireData();
      
      if(ptr == &stdDisplay) // default stats
      {
        Print("Symbol number: ", TradeRecord::getSymbolCount());
        for(int i = 0; i < TradeRecord::getSymbolCount(); i++)
        {
          Print(i, "] ", TradeRecord::getSymbol(i));
        }
      
        Print("Magic number: ", TradeRecord::getMagicCount());
        for(int i = 0; i < TradeRecord::getMagicCount(); i++)
        {
          Print(i, "] ", TradeRecord::getMagic(i));
        }
        
        Print("Filters: ", aggregator.getFilterTitles());
        
        Print("Selectors: ", selectorCount);
      }
      
      analyst.build();
      analyst.display(SortBy);
    
      delete analyst;
      delete aggregator;
      for(int i = 0; i < selectorCount; i++)
      {
        delete selectors[i];
      }
      for(int i = 0; i < ArraySize(filters); i++)
      {
        delete filters[i].getSelector();
        delete filters[i];
      }
    }
};

/*
   USAGE EXAMPLE

// Optional display class for custom output (if not provided - default logging is used);
// this example uses log, but other means can be used as well (charts, html rendering, csv file, etc)

class MyOLAPStats: public Display
{
  protected:
    void showVector(MetaCube *cube, const int &consts[], const SORT_BY sortby = SORT_BY_NONE) const
    {
      PairArray *result = NULL;
      cube.getVector(0, consts, result, sortby);
      if(CheckPointer(result) == POINTER_DYNAMIC)
      {
        ArrayPrint(result.array);
        delete result;
      }
    }

  public:
    virtual void display(MetaCube *cube, const SORT_BY sortby = SORT_BY_NONE, const bool identity = false) override
    {
      int consts[];
      const int n = cube.getDimension(); // selectorCount
      ArrayResize(consts, n);
      ArrayInitialize(consts, 0);
      
      if(n == 2)
      {
        const int i = 1;
        int m = cube.getDimensionRange(i);
        for(int j = 0; j < m; j++)
        {
          Print(cube.getDimensionTitle(i), " ", cube.getDimensionIndexLabel(i, j));
          consts[i] = j;
          
          showVector(cube, consts, sortby);
          
          consts[i] = 0;
        }
      }
      else if(n == 1)
      {
        Print(cube.getDimensionTitle(0));
        showVector(cube, consts, sortby);
      }
      else
      {
        Print("3D data output is not supported");
      }
    }
};

void OnDeinit(const int)
{
  Print("OLAP STATS");
  OLAPStats stats(SELECTOR_SYMBOL, FIELD_NONE, SELECTOR_PROFITABLE); // choose selectors and fields as required
  stats.setAggregator(AGGREGATOR_COUNT); // choose aggregator, could be e.g.: stats.setAggregator(AGGREGATOR_PROFITFACTOR, FIELD_PROFIT_POINTS);
  stats.setSorting(SORT_BY_VALUE_DESCENDING); // optionally choose sorting order
  MyOLAPStats callback;     // optional
  stats.process(&callback); // could be just stats.process();
}

*/
